@@ -4,7 +4,7 @@ CacheTools Utilities
 This code is public domain.
 """
 
-from typing import Optional, Callable, Dict, List, Set, Any, Union, MutableMapping
+from typing import Dict, Any, Union, MutableMapping
 
 import cachetools
 import json
@@ -12,22 +12,24 @@ import json
 import pkg_resources as pkg  # type: ignore
 __version__ = pkg.require("CacheToolsUtils")[0].version
 
+import logging
+log = logging.getLogger(__name__)
+
 
 #
 # UTILS
 #
 
-
 class MutMapMix:
     """Convenient MutableMapping Mixin, forward to _cache."""
 
-    def __getitem__(self, key):
+    def __getitem__(self, key):  # pragma: no cover
         return self._cache.__getitem__(key)
 
-    def __setitem__(self, key, val):
+    def __setitem__(self, key, val):  # pragma: no cover
         return self._cache.__setitem__(key, val)
 
-    def __delitem__(self, key):
+    def __delitem__(self, key):  # pragma: no cover
         return self._cache.__delitem__(key)
 
     def __len__(self):
@@ -40,9 +42,6 @@ class MutMapMix:
 class KeyMutMapMix(MutMapMix):
     """Convenient MutableMapping Mixin with a key filter, forward to _cache."""
 
-    def _key(self, key):
-        return key
-
     def __getitem__(self, key):
         return self._cache.__getitem__(self._key(key))
 
@@ -73,14 +72,16 @@ class StatsCache(MutMapMix, MutableMapping):
     """Cache class to keep stats."""
 
     def __init__(self, cache: MutableMapping):
-        self._reads, self._writes, self._dels, self._hits = 0, 0, 0, 0
         self._cache = cache
+        self.reset()
 
     def hits(self):
         return float(self._hits) / float(max(self._reads, 1))
 
+    def reset(self):
+        self._reads, self._writes, self._dels, self._hits = 0, 0, 0, 0
+
     def __getitem__(self, key):
-        # log.debug(f"get: {key} {key in self._cache}")
         self._reads += 1
         res = self._cache.__getitem__(key)
         self._hits += 1
@@ -98,11 +99,61 @@ class StatsCache(MutMapMix, MutableMapping):
         return self._cache.clear()
 
 
+class TwoLevelCache(MutMapMix, MutableMapping):
+    """Two-Level Cache class for CacheTools."""
+
+    def __init__(self, cache: MutableMapping, cache2: MutableMapping):
+        self._cache = cache
+        self._cache2 = cache2
+
+    def __getitem__(self, key):
+        try:
+            return self._cache.__getitem__(key)
+        except KeyError:
+            val = self._cache2.__getitem__(key)
+            self._cache.__setitem__(key, val)
+            return val
+
+    def __setitem__(self, key, val):
+        self._cache2.__setitem__(key, val)
+        return self._cache.__setitem__(key, val)
+
+    def __delitem__(self, key):
+        try:
+            self._cache2.__delitem__(key)
+        except KeyError:
+            pass
+        return self._cache.__delitem__(key)
+
+    def clear(self):
+        return self._cache.clear()
+
+
+def cacheMethods(cache: MutableMapping, obj: Any, funs: Dict[str, str]):
+    """Cache some object methods."""
+    for fun, prefix in funs.items():
+        assert hasattr(obj, fun), f"cannot cache missing method {fun} on {obj}"
+        f = getattr(obj, fun)
+        while hasattr(f, "__wrapped__"):
+            f = f.__wrapped__
+        setattr(obj, fun,
+                cachetools.cached(cache=PrefixedCache(cache, prefix))(f))
+
+
+def cacheFunctions(cache: MutableMapping, globs: MutableMapping[str, Any],
+                   funs: Dict[str, str]):
+    for fun, prefix in funs.items():
+        assert fun in globs, "caching functions: {fun} not found"
+        f = globs[fun]
+        while hasattr(f, "__wrapped__"):
+            f = f.__wrapped__
+        globs[fun] = cachetools.cached(cache=PrefixedCache(cache, prefix))(f)
+
+
 #
 # MEMCACHED
 #
 
-
 class JsonSerde:
     """JSON serialize/deserialize for MemCached."""
 
@@ -119,7 +170,7 @@ class JsonSerde:
             return value.decode("utf-8")
         elif flags == 2:
             return json.loads(value.decode("utf-8"))
-        else:
+        else:  # pragma: no cover
             raise Exception("Unknown serialization format")
 
 
@@ -142,7 +193,7 @@ class MemCached(KeyMutMapMix, MutableMapping):
     def stats(self):
         return self._cache.stats()
 
-    def clear(self):
+    def clear(self):  # pragma: no cover
         return self._cache.flush_all()
 
 
@@ -182,7 +233,7 @@ class RedisCache(MutableMapping):
         self._cache = cache
         self._ttl = ttl
 
-    def clear(self):
+    def clear(self):  # pragma: no cover
         return self._cache.flushdb()
 
     def _serialize(self, s):
@@ -202,7 +253,8 @@ class RedisCache(MutableMapping):
             raise KeyError()
 
     def __setitem__(self, index, value):
-        return self._cache.set(self._key(index), self._serialize(value), ex=self._ttl)
+        return self._cache.set(self._key(index), self._serialize(value),
+                               ex=self._ttl)
 
     def __delitem__(self, index):
         return self._cache.delete(self._key(index))
@@ -213,7 +265,8 @@ class RedisCache(MutableMapping):
     def __iter__(self):
         raise Exception("not implemented yet")
 
-    # also forward Redis set/get/delele
+    # also forward Redis set/get/delete
+    # FIXME should it pass through the key filter?
     def info(self, *args, **kwargs):
         return self._cache.info(*args, **kwargs)
 
@@ -238,7 +291,8 @@ class PrefixedRedisCache(RedisCache):
         self._prefix = prefix
 
     def _key(self, key):
-        return self._prefix + json.dumps(key)
+        # or self._cache._key(prefix + key)?
+        return self._prefix + str(key)
 
 
 class StatsRedisCache(RedisCache):
@@ -249,4 +303,5 @@ class StatsRedisCache(RedisCache):
 
     def hits(self):
         stats = self.info(section="stats")
-        return float(stats["keyspace_hits"]) / (stats["keyspace_hits"] + stats["keyspace_misses"])
+        return float(stats["keyspace_hits"]) / \
+            (stats["keyspace_hits"] + stats["keyspace_misses"])
